; Fast assembler implementation of Bresenham's line-drawing algorithm
; for the Tseng Labs SuperVGA.  Works in mode 56 (decimal).
; Turbo Pascal 6.0 far-callable.
;
; *********************************************************************
;  Pascal-compatible line-drawing entry point at _Line.               *
;  Called from Turbo Pascal with:                                     *
;        _Line(X0,Y0,X1,Y1,Color);                                    *
; *********************************************************************
;

_TEXT   Segment Byte Public 'CODE'
        Assume CS: _TEXT

;
; Equates.
;
VGA_SCREEN_WIDTH_IN_BYTES EQU 320             ; Memory offset from start of
                                              ;  one row to start of next
                                              ;  in display memory.
VGA_SCREEN_SEGMENT        EQU 0A000h          ; Display-memory segment

;
; Stack frame.
;
VGALineParms   Struc
       DW      ?              ; Pushed BP
       DD      ?              ; Pushed return address
Color  DW      ?              ; Starting X coordinate of line
Y1     DW      ?              ; Starting Y coordinate of line
X1     DW      ?              ; Ending X coordinate of line
Y0     DW      ?              ; Ending Y coordinate of line
X0     DW      ?              ; Color of line
VGALineParms   EndS

; *********************************************************************
;  Line-drawing macros.                                               *
; *********************************************************************

;
; Macro to loop through length of line, drawing each pixel in turn.
; Used for case of |DeltaX| >= |DeltaY|.
; |DeltaX|+1 points are drawn.
;
; Input:
;       MOVE_LEFT: 1 if DeltaX < 0, 0 else
;       AL: Line color
;       BX: DeltaX (X distance between start and end poimts)
;       SI: DeltaY (Y distance between start and end points)
;       ES:DI: Display-memory address of byte containing initial pixel
;
; Output: none
;
LINE1   Macro   MOVE_LEFT
        Local   LineLoop, MoveXCoord, Line1End, MoveToNextByte
        MOV     CX,BX         ; # of pixels in line
        JCXZ    Line1End      ; Done if there are no more pixels
                              ;  (there's always at least one pixel
                              ;  at the start location)
        SHL     SI,1          ; DeltaY * 2
        MOV     BP,SI         ; Error term
        SUB     BP,BX         ; Error term starts at DeltaY * 2 - DeltaX
        SHL     BX,1          ; DeltaX * 2
        SUB     SI,BX         ; DeltaY * 2 - DeltaX * 2 (used in loop)
        ADD     BX,SI         ; DeltaY * 2 (used in loop)
        MOV     AH,AL
LineLoop:
;
; See if it's time to advance the Y coordinate yet.
;
        AND     BP,BP         ; See if error term is negative
        JS      MoveXCoord    ; Yes, stay at the same Y coordinate
;
; Advance the Y coordinate, first writing all pixels in the current byte.
;
        MOV     BYTE PTR [DI],AL
        ADD     DI,VGA_SCREEN_WIDTH_IN_BYTES
        ADD     BP,SI         ; Adjust error term back down
        JMP     SHORT MoveToNextByte
;
; Adjust display memory address.
;
MoveXCoord:
        ADD     BP,BX         ; Increment error term & keep same Y coordinate
        MOV     BYTE PTR [DI],AL
MoveToNextByte:
If MOVE_LEFT
        DEC     DI            ; Next pixel is in byte to left
Else
        INC     DI            ; Next pixel is in byte to right
EndIf
        LOOP    LineLoop
Line1End:
        MOV     BYTE PTR [DI],AL
        EndM

;
; Macro to loop through length of line, drawing each pixel in turn.
; Used for case of DeltaX < DeltaY.
; |DeltaY|+1 points are drawn.
;
; Input:
;       MOVE_LEFT: 1 if DeltaX < 0, 0 else
;       AL: Line color
;       BX: |DeltaX| (X distance between start and end points)
;       SI: DeltaY   (Y distance between start and end points)
;       ES:DI: Display-memory address of byte containing initial pixel
;
; Output: none
;
Line2   Macro   MOVE_LEFT
        Local   LineLoop, MoveYCoord, ETermAction, Line2End
        MOV     CX,SI         ; # of pixels in line
        SHL     BX,1          ; DeltaX * 2
        MOV     BP,BX         ; Error term
        SUB     BP,SI         ; Error term starts at DeltaX * 2 - DeltaY
        SHL     SI,1          ; DeltaY * 2
        SUB     BX,SI         ; DeltaX * 2 - DeltaY * 2 (used in loop)
        ADD     SI,BX         ; DeltaX * 2 (used in loop)
;
; Draw initial pixel.
;
        MOV     BYTE PTR [DI],AL
        JCXZ    Line2End
LineLoop:
;
; See if it's time to advance the X coordinate yet.
;
        AND     BP,BP         ; See if error term is negative
        JNS     ETermAction   ; No, advance X coordinate
        ADD     BP,SI         ; Increment error term & keep same X coordinate
        JMP     SHORT MoveYCoord
ETermAction:
;
; Adjust display memory address.
;
If MOVE_LEFT
        DEC     DI
Else
        INC     DI
EndIf
        ADD     BP,BX         ; Adjust error term back down
;
; Advance the Y coordinate.
;
MoveYCoord:
        ADD     DI,VGA_SCREEN_WIDTH_IN_BYTES
;
; Write the next pixel.
;
        MOV     BYTE PTR [DI],AL
        LOOP    LineLoop
Line2End:
        EndM

; *********************************************************************
;  Line-drawing routine.                                              *
; *********************************************************************

        Public  _Line
_Line   Proc    Far
        PUSH    BP
        MOV     BP,SP
        PUSH    SI            ; Preserve register variables
        PUSH    DI
        PUSH    DS
;
; Point DS to display memory.
;
        MOV     AX,VGA_SCREEN_SEGMENT
        MOV     DS,AX
;
; Get DeltaY.
;
        MOV     SI,[BP+Y1]    ; Line Y start
        MOV     AX,[BP+Y0]    ; Line Y end, used later in calculating
                              ; the start address
        SUB     SI,AX         ; Calculate DeltaY
        JNS     CalcStartAddress
;
; DeltaY is negative -- swap coordinates so we're always working with a
; positive DeltaY.
;
        MOV     AX,[BP+Y1]    ; Set line start to Y1, for use in
                              ; calculating the start address
        MOV     DX,[BP+X0]
        XCHG    DX,[BP+X1]
        MOV     [BP+X0],DX    ; Swap X coordinates
        NEG     SI            ; Convert to positive DeltaY
;
; Calculate the starting address in display memory of the line.
;
CalcStartAddress:
        MOV     CX,VGA_SCREEN_WIDTH_IN_BYTES
        MUL     CX
        MOV     DX,[BP+X0]
        ADD     AX,DX
        MOV     DI,AX
        MOV     AX,[BP+Color]
;
; Calculate DeltaX.
;
        MOV     BX,[BP+X1]
        SUB     BX,[BP+X0]
;
; Handle correct one of four octants.
;
        JS      NegDeltaX
        CMP     BX,SI
        JB      Octant1
;
; DeltaX >= DeltaY >= 0.
;
        LINE1   0
        JMP     VGALineDone
;
; DeltaY > DeltaX >= 0.
;
Octant1:
        LINE2   0
        JMP     SHORT VGALineDone
;
NegDeltaX:
        NEG     BX            ; |DeltaX|
        CMP     BX,SI
        JB      Octant2
;
; |DeltaX| >= DeltaY and DeltaX < 0.
;
        LINE1   1
        JMP     SHORT VGALineDone
;
; |DeltaX| < DeltaY and DeltaX < 0.
;
Octant2:
        LINE2   1
;
VGALineDone:
;
; Pop registers and exit.
;
        POP     DS
        POP     DI
        POP     SI
        POP     BP
        RET
_Line   EndP

_TEXT   EndS
        End
